@@ -1,7 +1,7 @@
 %%
 %% %CopyrightBegin%
 %%
-%% Copyright Ericsson AB 2010-2023. All Rights Reserved.
+%% Copyright Ericsson AB 2010-2024. All Rights Reserved.
 %%
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -21,8 +21,9 @@
 -export([all/0, suite/0,groups/0,init_per_suite/1, end_per_suite/1,
          init_per_testcase/2, end_per_testcase/2,
          init_per_group/2,end_per_group/2]).
--export([normal/1, type_completion/1, quoted_fun/1, quoted_module/1, quoted_both/1, erl_1152/1, get_coverage/1,
-         check_trailing/1, unicode/1, filename_completion/1, binding_completion/1, record_completion/1,
+-export([normal/1, type_completion/1, quoted_fun/1, quoted_module/1, quoted_both/1,
+         invalid_module/1, erl_1152/1, get_coverage/1, check_trailing/1, unicode/1,
+         filename_completion/1, binding_completion/1, record_completion/1, no_completion/1,
          map_completion/1, function_parameter_completion/1, fun_completion/1]).
 -record(a_record,
         {a_field   :: atom1 | atom2 | btom | 'my atom' | {atom3, {atom4, non_neg_integer()}} | 'undefined',
@@ -50,8 +51,8 @@ suite() ->
 all() ->
     [normal, filename_completion, binding_completion, get_coverage, type_completion, 
      record_completion, fun_completion, map_completion, function_parameter_completion,
-     quoted_fun, quoted_module, quoted_both, erl_1152, check_trailing,
-     unicode].
+     no_completion, quoted_fun, quoted_module, quoted_both, erl_1152, check_trailing,
+     invalid_module, unicode].
 
 groups() ->
     [].
@@ -108,13 +109,14 @@ type_completion(_Config) ->
     ct:timetrap({minutes, 20}),
     {Time,_} = timer:tc(fun() -> do_expand("erl_pp:expr(") end),
     case Time of
-        Time when Time > 2600000 -> {skip, "Expansion too slow on this machine"};
+        Time when Time > 5000000 ->
+            {skip, lists:flatten(io_lib:format("Expansion too slow (~p) on this machine",[Time]))};
         _ ->
             parallelforeach(
             fun(Mod) ->
                     Exports = edlin_expand:get_exports(Mod),
                     [try
-                        Str = io_lib:write_atom(Mod) ++ ":" ++ atom_to_list(Func) ++ "(",
+                        Str = io_lib:write_atom(Mod) ++ ":" ++ io_lib:write_atom(Func) ++ "(",
                         do_expand(Str)
                     catch E:R:ST ->
                             erlang:raise(E, {R, Mod, Func}, ST)
@@ -173,6 +175,12 @@ filename_completion(Config) ->
     file:set_cwd(CWD),
     R.
 
+no_completion(_Config) ->
+    %% No autocompletion, and no crashes
+    {no, _, _} = do_expand("[{"),
+    {no, _, _} = do_expand("a [{"),
+    {no, _, _} = do_expand("a [{{"),
+    ok.
 record_completion(_Config) ->
     %% test record completion for loaded records
     %% test record field name completion
@@ -197,12 +205,12 @@ record_completion(_Config) ->
                {"c_field",[{ending,"="}]},
                {"d_field",[{ending,"="}]}],
               options:=[highlight_all]}]} = do_expand("#a_record{a_field={atom3,b},"),
-    %% test match argument and closing parenthesis completion
-    {yes,", ",[]} = do_expand("#a_record{a_field={atom3"),
+    %% test match argument
+    {yes,_,[]} = do_expand("#a_record{a_field={atom3"),
     {no,[],[#{title:="types",elems:=[{"{atom4, ...}",[]}],options:=[{hide,title}]}]} = do_expand("#a_record{a_field={atom3,"),
     {no,[],[#{title:="types",elems:=[{"integer() >= 0",[]}],options:=[{hide,title}]}]} = do_expand("#a_record{a_field={atom3,{atom4, "),
-    {yes,"}",_} = do_expand("#a_record{a_field={atom3,{atom4, 1"),
-    {yes,"}",_} = do_expand("#a_record{a_field={atom3,{atom4, 1}"),
+    {no,_,_} = do_expand("#a_record{a_field={atom3,{atom4, 1"),
+    {no,_,_} = do_expand("#a_record{a_field={atom3,{atom4, 1}"),
     ok.
 
 fun_completion(_Config) ->
@@ -246,11 +254,15 @@ function_parameter_completion(Config) ->
     {no, [], [#{elems:=[#{elems:=[#{title:="types",elems:=[{"integer()",[]}]}]}]}]} = do_expand("complete_function_parameter:a_fun_name("),
     {no, [], [#{elems:=[#{elems:=[#{elems:=[{"integer()",[]}]}]}]}]} = do_expand("complete_function_parameter:a_fun_name(1,"),
     {no, [], [#{elems:=[#{elems:=[#{elems:=[{"integer()",[]}]}]}]}]}  = do_expand("complete_function_parameter : a_fun_name ( 1 , "),
-    {yes, ")", []} = do_expand("complete_function_parameter:a_fun_name(1,2"),
+    {no, [], []} = do_expand("complete_function_parameter:a_fun_name(1,2"),
     {no, [], []} = do_expand("complete_function_parameter:a_fun_name(1,2,"),
     {no, [], [#{elems:=[#{elems:=[#{elems:=[{"any()",[]},{"[any() | [Deeplist]]",[]}]}]}]}]} = do_expand("complete_function_parameter:a_deeplist_fun("),
     {no,[],[#{title:="typespecs",
                 elems:=[#{title:=
+                                "complete_function_parameter:multi_arity_fun()",
+                            options:=[],
+                            elems:=[{")",[]}]},
+                        #{title:=
                                "complete_function_parameter:multi_arity_fun(T1)",
                           elems:=[#{title:="types",
                                     elems:=[{"integer()",[]}],
@@ -261,13 +273,10 @@ function_parameter_completion(Config) ->
                             elems:=[#{title:="types",
                                     elems:=[{"integer()",[]}],
                                     options:=[{hide,title}]}],
-                            options:=[{highlight_param,1}]},
-                            #{title:=
-                                "complete_function_parameter:multi_arity_fun()",
-                            options:=[],
-                            elems:=[")"]}],
+                            options:=[{highlight_param,1}]}],
                 options:=[highlight_all]}]} = do_expand("complete_function_parameter:multi_arity_fun("),
     {no, [], [#{elems:=[#{elems:=[#{elems:=[{"true",[]},{"false",[]}]}]}]}]} = do_expand("complete_function_parameter:multi_arity_fun(1,"),
+    {no, [], []} = do_expand("complete_function_parameter:multi_arity_fun(["),
     {no,[],
         [#{elems :=
             [#{elems :=
@@ -354,6 +363,10 @@ get_coverage(Config) ->
     do_expand("complete_function_parameter:map_parameter_function(#{}, "),
     do_expand("complete_function_parameter:map_parameter_function(#{V=>1}, "),
     do_expand("complete_function_parameter:map_parameter_function(#{a=>V}, "),
+    do_expand("complete_function_parameter:map_variable_parameter_function(#{"),
+    do_expand("complete_function_parameter:map_variable_parameter_function(#{a"),
+    do_expand("complete_function_parameter:map_variable_parameter_function(#{a => "),
+    do_expand("complete_function_parameter:map_variable_parameter_function(#{a => a"),
     do_expand("complete_function_parameter:tuple_parameter_function({a,b}, "),
     do_expand("complete_function_parameter:tuple_parameter_function({a,V}, "),
     do_expand("complete_function_parameter:list_parameter_function([], "),
@@ -509,6 +522,9 @@ get_coverage(Config) ->
     {_, _, M10} = edlin_expand:expand("ssh:connect({},["),
     do_format(M10),
     lists:flatten(edlin_expand:format_matches(M10, 20)),
+    %% Test that we are not filtering duplicates bit with different case or different string lengths
+    {yes,"e", M11} = do_expand("complete_function_parameter:cas"),
+    "\e[;1;4mfunctions\e[0m\ncaseSensitiveFunction(        casesensitivefunction(        \ncaseSensitiveFunctionName(\n" = do_format(M11),
     ok.
 
 %% Normal module name, some function names using quoted atoms.
@@ -555,6 +571,23 @@ quoted_module(Config) when is_list(Config) ->
                                                 {"a_less_fun_name",_}]}]} = do_expand("'ExpandTestCaps':a_"),
     ok.
 
+%% Test that expansion does not break when module/function is invalid
+invalid_module(Config) when is_list(Config) ->
+
+    {no, "", []} = do_expand("0"),
+    {no, "", []} = do_expand("0:"),
+    {no, "", []} = do_expand("0:a"),
+    {no, "", []} = do_expand("0:a("),
+    {no, "", []} = do_expand("lists:0"),
+    {no, "", []} = do_expand("lists:0("),
+    {no, "", []} = do_expand("fun 0"),
+    {no, "", []} = do_expand("fun 0:"),
+    {no, "", []} = do_expand("fun 0:a"),
+    {no, "", []} = do_expand("fun 0:a/"),
+    {no, "", []} = do_expand("fun lists:0"),
+    {no, "", []} = do_expand("fun lists:0/"),
+    ok.
+
 quoted_both(Config) when is_list(Config) ->
     {module,'ExpandTestCaps'} = compile_and_load(Config,'ExpandTestCaps'),
     {module,'ExpandTestCaps1'} = compile_and_load(Config,'ExpandTestCaps1'),
@@ -607,7 +640,7 @@ unicode(Config) when is_list(Config) ->
     ok.
 
 do_expand(String) ->
-    erlang:display(String),
+    io:format("~ts", [String]),
     Bs = [
           {'Binding', 0},
           {'MapBinding', #{a_key=>0, b_key=>1, c_key=>2}},
