@@ -1,7 +1,7 @@
 %%
 %% %CopyrightBegin%
 %%
-%% Copyright Ericsson AB 1998-2023. All Rights Reserved.
+%% Copyright Ericsson AB 1998-2024. All Rights Reserved.
 %%
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -30,8 +30,10 @@
          otp_11728/1, encoding/1, extends/1,  function_macro/1,
 	 test_error/1, test_warning/1, otp_14285/1,
 	 test_if/1,source_name/1,otp_16978/1,otp_16824/1,scan_file/1,file_macro/1,
-         triple_quotes_warning/1,
-   deterministic_include/1, nondeterministic_include/1]).
+         deterministic_include/1, nondeterministic_include/1,
+         gh_8268/1,
+         moduledoc_include/1
+        ]).
 
 -export([epp_parse_erl_form/2]).
 
@@ -74,8 +76,9 @@ all() ->
      otp_8665, otp_8911, otp_10302, otp_10820, otp_11728,
      encoding, extends, function_macro, test_error, test_warning,
      otp_14285, test_if, source_name, otp_16978, otp_16824, scan_file, file_macro,
-     triple_quotes_warning,
-     deterministic_include, nondeterministic_include].
+     deterministic_include, nondeterministic_include,
+     gh_8268,
+     moduledoc_include].
 
 groups() ->
     [{upcase_mac, [], [upcase_mac_1, upcase_mac_2]},
@@ -129,6 +132,48 @@ file_macro(Config) when is_list(Config) ->
     "Other source" = FileA = FileB,
     ok.
 
+moduledoc_include(Config) when is_list(Config) ->
+    PrivDir = proplists:get_value(priv_dir, Config),
+    ModuleFileContent = <<"-module(moduledoc).
+
+                           -moduledoc {file, \"README.md\"}.
+
+                           -export([]).
+                          ">>,
+    DocFileContent = <<"# README
+
+                        This file is a test
+                       ">>,
+    CreateFile = fun (Dir, File, Content) ->
+                     Dirname = filename:join([PrivDir, Dir]),
+                     ok = create_dir(Dirname),
+                     Filename = filename:join([Dirname, File]),
+                     ok = file:write_file(Filename, Content),
+                     Filename
+                 end,
+
+    %% positive test: checks that all works as expected
+    ModuleName = CreateFile("module_attr", "moduledoc.erl", ModuleFileContent),
+    DocName = CreateFile("module_attr", "README.md", DocFileContent),
+    {ok, List} = epp:parse_file(ModuleName, []),
+    {attribute, _, moduledoc, ModuleDoc} = lists:keyfind(moduledoc, 3, List),
+    ?assertEqual({ok, unicode:characters_to_binary(ModuleDoc)}, file:read_file(DocName)),
+
+    %% negative test: checks that we produce an expected error
+    ModuleErrContent = binary:replace(ModuleFileContent, <<"README">>, <<"NotExistingFile">>),
+    ModuleErrName = CreateFile("module_attr", "moduledoc_err.erl", ModuleErrContent),
+    {ok, ListErr} = epp:parse_file(ModuleErrName, []),
+    {error,{_,epp,{moduledoc,file, "NotExistingFile.md"}}} = lists:keyfind(error, 1, ListErr),
+
+    ok.
+
+create_dir(Dir) ->
+    case file:make_dir(Dir) of
+        ok -> ok;
+        {error, eexist} -> ok;
+        _ -> error
+    end.
+
 deterministic_include(Config) when is_list(Config) ->
     DataDir = proplists:get_value(data_dir, Config),
     File = filename:join(DataDir, "deterministic_include.erl"),
@@ -932,11 +977,12 @@ scan_file(Config) when is_list(Config) ->
     [FileForm1, ModuleForm, ExportForm,
      FileForm2, FileForm3, FunctionForm,
      {eof,_}] = Toks,
-    [{'-',_}, {atom,_,file}, {'(',_} | _ ] = FileForm1,
+    [{'-',_}, {atom,_,file}, {'(',_} | _ ]   = FileForm1,
     [{'-',_}, {atom,_,module}, {'(',_} | _ ] = ModuleForm,
     [{'-',_}, {atom,_,export}, {'(',_} | _ ] = ExportForm,
-    [{'-',_}, {atom,_,file}, {'(',_} | _ ] = FileForm2,
-    [{'-',_}, {atom,_,file}, {'(',_} | _ ] = FileForm3,
+    [{'-',_}, {atom,_,file}, {'(',_} | _ ]   = FileForm2,
+    [{'-',_}, {atom,_,file}, {'(',_} | _ ]   = FileForm3,
+    [{atom,_,ok}, {'(',_} | _]               = FunctionForm,
     ok.
 
 macs(Epp) ->
@@ -1256,8 +1302,15 @@ test_if(Config) ->
 	  {if_8c,
 	   <<"-if(?foo).\n"                     %Undefined symbol.
 	     "-endif.\n">>,
-	   {errors,[{{1,25},epp,{undefined,foo,none}}],[]}}
+	   {errors,[{{1,25},epp,{undefined,foo,none}}],[]}},
 
+	  {if_9c,
+	   <<"-if(not_builtin()).\n"
+	     "a bug.\n"
+	     "-else.\n"
+	     "t() -> ok.\n"
+	     "-endif.\n">>,
+	   {errors,[{{1,21},epp,{bad,'if'}}],[]}}
 	 ],
     [] = compile(Config, Cs),
 
@@ -1323,14 +1376,6 @@ test_if(Config) ->
            ok},
 
 	  {if_7,
-	   <<"-if(not_builtin()).\n"
-	     "a bug.\n"
-	     "-else.\n"
-	     "t() -> ok.\n"
-	     "-endif.\n">>,
-           ok},
-
-	  {if_8,
 	   <<"-if(42).\n"			%Not boolean.
 	     "a bug.\n"
 	     "-else.\n"
@@ -2057,73 +2102,26 @@ otp_16824(Config) when is_list(Config) ->
     [] = compile(Config, Cs),
     ok.
 
-triple_quotes_warning(Config) when is_list(Config) ->
-    Cs1 =
-        [{triple_quotes_warning_1,
-          <<"\n-doc \"foo\".\n">>,
-          []},
-         {triple_quotes_warning_2,
-          <<"\n-doc \"\" \"foo\" \"\".\n">>,
-          []},
-         {triple_quotes_warning_3,
-          <<"\n"
-            "-doc \"\"\"\n"
-            "    foo\n"
-            "    \"\"\".\n">>,
-          {warnings,[{{2,6},epp,tqstring}]}},
-         {triple_quotes_warning_4,
-          <<"\n"
-            "-doc  \"\"\"\"\n"
-            "      \"\"\"\".\n">>,
-          {warnings,
-           [{{2,7},epp,tqstring},
-            {{3,7},epp,tqstring}]}},
-         {triple_quotes_warning_5,
-          <<"\n"
-            "-doc   \"\"\"\"\"\n"
-            "       foo\n"
-            "       \"\"\"\"\".\n">>,
-          {warnings,
-           [{{2,8},epp,tqstring},
-            {{4,9},epp,tqstring}]}}
-        ],
-    [] = compile(Config, Cs1),
-
-    Cs2 =
-        [{triple_quotes_warning_10,
-          <<"\n"
-            "-export([foo/0]).\n"
-            "foo() ->\n"
-            "    \"\"\"\n"
-            "    bar\n"
-            "    \"\"\".\n">>,
-          {warnings,[{{4,5},epp,tqstring}]}},
-         {triple_quotes_warning_11,
-          <<"\n"
-            "-export([foo/0]).\n"
-            "foo() ->\n"
-            "    \"\"\"\"\n"
-            "    ++ lists:duplicate(4, $x) ++\n"
-            "    \"\"\"\".\n">>,
-          {warnings,
-           [{{4,5},epp,tqstring},
-            {{6,5},epp,tqstring}]}},
-         {triple_quotes_warning_12,
-          <<"\n"
-            "-export([foo/0]).\n"
-            "foo() ->\n"
-            "    \"\"\"\"\"\n"
-            "    bar\n"
-            "    \"\"\"\"\".\n">>,
-          {warnings,
-           [{{4,5},epp,tqstring},
-            {{6,6},epp,tqstring}]}} ],
-    [] = compile(Config, Cs2),
-
+gh_8268(Config) ->
+    Ts = [{circular_1,
+           <<"-define(LOG(Tag, Code), io:format(\"~s\", [Tag]), Code).
+             more_work() -> ok.
+             some_work() -> ok.
+             t() ->
+                ?LOG(work,
+                     begin
+                        maybe
+                           ok ?= some_work()
+                        end,
+                        more_work()
+                     end).
+             ">>,
+           [{feature,maybe_expr,enable}],
+           ok}],
+    [] = run(Config, Ts),
     ok.
 
 
-
 %% Start location is 1.
 check(Config, Tests) ->
     eval_tests(Config, fun check_test/3, Tests).
@@ -2154,8 +2152,6 @@ eval_tests(Config, Fun, Tests) ->
                         case E of
                             {errors, Errors} ->
                                 call_format_error(Errors);
-                            {warnings, Errors} ->
-                                call_format_error(Errors);
                             _ ->
                                 ok
                         end,
